#include "dg_node.h"
#include "det.h"
#include <string>
#include <map>
#include <set>
#include <unordered_set>
#include <algorithm>
#include <fstream>
#include <sstream>
#include <iostream>
#include <vector>
#include <queue> 
#include <curl/curl.h>
#include <curl/easy.h>
#include <ctype.h>
#include "debug.h"

/*

	det.cc contains all functions related to IoT detection phase of IoTSTEED

*/


using namespace std;


/*
        convert string to lower case
*/

void str2lower(string& data){
        std::transform(data.begin(), data.end(), data.begin(), ::tolower);
}

/*
	replace multiple white spaces in string as one white space	
*/
void compressSpace(string& s) {
	string::size_type pos;
	while((pos = s.find("  ")) != string::npos)
	{
	  s.replace(pos, 2, " ");
	}
}

/*
        reads a list of mftr to mrlv mapping, build DG_Node for each new 
        mftr and mrlv, connect them and store them in a map<string, DG_Node>;
	this map<string, DG_Node> is our directed graph of prior knowledge;
	special mrlv: - stands for mftr has no mrlv; null stands for mftr has special non-existent mrlv
*/
map<string, DG_Node> createDGfromCSV(const char* fn){
	map<string, DG_Node> name2node;
	typedef map<string, DG_Node>::iterator mitr;          
	ifstream infile(fn);
	char del=',';
	string line, mftr, mrlv;

	while(getline(infile,line))
	{	
		stringstream ss(line);
		getline(ss, mftr, del);
	        getline(ss, mrlv, del);

		if(mftr.compare("mftr") == 0) continue; //skip header line

                /*
                         we store mftr and mrlv internally:
                         (1) lower case
                         (2) remove any special characters, leaving only alphabets, numerics and space
                         (3) remove any extra space; i.e. single space only
                */
                str2lower(mftr); //(1) lower case
                mftr.erase(remove_if(mftr.begin(), mftr.end(), [](char c) { return ! (isalnum(c) || isspace(c)) ; }), mftr.end()); // (2) leaving only alpha, num, space
                compressSpace(mftr); //(3) remove any extra space

		if(name2node.find(mftr)==name2node.end()) // if new mftr, create and store node
		{
			DG_Node mftr_node(mftr,true);
			name2node.insert(pair<string, DG_Node>(mftr,mftr_node));
		}
		else if ((name2node.find(mftr)->second).isMftr()==false) // if an exsiting mrlv turn out to be also a mftr, change it to mftr
		{
			(name2node.find(mftr)->second).changeType(true);
		}

		if(mrlv.compare("-") == 0) continue;// skip non-existent mrlv

                /*
                         we store mftr and mrlv internally:
                         (1) lower case
                         (2) remove any special characters, leaving only alphabets, numerics and space
                         (3) remove any extra space; i.e. single space only
                */
                str2lower(mrlv); //(1) lower case
                mrlv.erase(remove_if(mrlv.begin(), mrlv.end(), [](char c) { return ! (isalnum(c) || isspace(c)) ; }), mrlv.end()); // (2) leaving only alpha, num, space
                compressSpace(mrlv); //(3) remove any extra space

                if(name2node.find(mrlv)==name2node.end()) // if new mrlv, create and store node
                {
                        DG_Node mrlv_node(mrlv,false);
                        name2node.insert(pair<string, DG_Node>(mrlv,mrlv_node));
                }

		//link mrlv to mftr
		mitr i_mftr=name2node.find(mftr);
		mitr i_mrlv=name2node.find(mrlv);
		if(i_mftr==name2node.end() || i_mrlv==name2node.end() ) // this shouldn't happen
		{
			cerr << "[ERROR] non-existent mrlv " << mrlv <<
		        " or mftr " << mftr << " in readMftrMrlv()." << endl;
                        exit(-1);
		}
		(i_mrlv->second).addNextNode(i_mftr->second);
	}
	return name2node;
}

/*
	Look up orgnaization name from MAC address, with API from macaddress.io

*/

size_t write_data(void *ptr, size_t size, size_t nmemb, void *stream) {
    string data((const char*) ptr, (size_t) size * nmemb);
    *((stringstream*) stream) << data << endl;
    return size * nmemb;
}

string maclookup(const string& mac_addr, const string& apikey){
#ifdef DEBUG
        clog << "##Looking up MAC " << mac_addr;
#endif
    string host =        "api.macaddress.io";
    string message_str = "https://"+host+"/v1?apikey="+apikey+"&output=vendor&search="+mac_addr;	
    CURL* curl=curl_easy_init();
    CURLcode res;
    stringstream out;

    curl_global_init(CURL_GLOBAL_DEFAULT);

    if (curl) {
        curl_easy_setopt(curl, CURLOPT_URL, message_str.c_str());
	curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1L);
    	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, write_data);
    	curl_easy_setopt(curl, CURLOPT_WRITEDATA, &out);
    }else{
        cerr << "[ERROR] can't create curl object when connecting to" << host <<
        " in maclookup()" << endl;
        exit(-1);
    }
	        
   res=curl_easy_perform(curl);
   if(res != CURLE_OK){
        cerr << "[ERROR] curl perform failed:" << curl_easy_strerror(res) <<
        " in maclookup()" << endl;
        curl_easy_cleanup(curl);
        exit(-1);
    }
   
    curl_global_cleanup();
    string ret=out.str();
    if (!ret.empty() && ret[ret.length()-1] == '\n') {
    	ret.erase(ret.length()-1);
    }   // remove potential trailing newline
	
    if (ret.empty()){
#ifdef DEBUG   
	clog <<"; Result=null" << endl;
#endif
	return "null";
    } else {
#ifdef DEBUG
	clog <<"; Result="<< ret << endl;
#endif
    	return ret;
    }
	
}


/*
        Look up DG_Node in DG (name2node) with orgnaization name; for a hit, return pointer to the macthing DG_Node, for a miss, return NULL pointer;

	Matching method: compare match lookup_result with every known mftr and see if any mftr name is a sub-sentence of match lookup_result
*/
const DG_Node* findNodeFromName(string& org, const map<string, DG_Node>& name2node){
#ifdef DEBUG
        clog << "##Searching DG_Node matching ";
#endif
        /*
                  convert org (the MAC lookup result) to our internal storage form of mftr and mrlv:
                  (1) lower case
                  (2) remove any special characters, leaving only alphabets, numerics and space
                  (3) remove any extra space; i.e. single space only
         */

        str2lower(org); //(1) lower case
        org.erase(remove_if(org.begin(), org.end(), [](char c) { return ! (isalnum(c) || isspace(c)) ; }), org.end()); // (2) leaving only alpha, num, space
        compressSpace(org); //(3) remove any extra space
#ifdef DEBUG
        clog << "\"" <<  org << "\"" << endl;
#endif

	for(map<string, DG_Node>::const_iterator iter=name2node.begin(); iter!=name2node.end(); iter++)
	{   
		
		if(org.find(iter->first) != string::npos) // if the known mftr (iter->first) is a substr of org, a possible hit
		{
			/*Test if the known mftr (iter->first) is a subsentence of org. If yes, then a hit */
			istringstream orgWds(org, istringstream::in);  //split org string into words
			istringstream mftrWds(iter->first, istringstream::in);  //split mftr string into words
        		string oWd; // tmp string to iterate words in org string
			string mWd; // tmp string to iterate words in mftr string
			bool start=false;

			mftrWds >> mWd;
        		while (orgWds >> oWd) //iterate words from org in cWd
        		{	
				if(oWd.compare(mWd)==0)
				{
					if(!start){start=true;}
					if(mftrWds.eof()){
#ifdef DEBUG
        					clog << "###Hit: found matching IoT mftr/mrlv node \"" << iter->first << "\" ==> new device is IoT" << endl;         
#endif
						return &(iter->second);

					} else {
						mftrWds >> mWd;
					}
				}
				else if (start)
			{
					break;
				}
			}
		}
	}
#ifdef DEBUG
      clog << "###Miss ==> new device is non-IoT" << endl;
#endif
	return NULL;
}
	

/*
	(Deprecated) Look up DG_Node in DG (name2node) with orgnaization name; for a hit, return pointer
 	to the macthing DG_Node, for a miss, return NULL pointer;

	Matching method: since MAC_lookup result is usually [prefix]+mftr_name+[suffix], we slide a window of size n
  	through the MAC_lookup result, in hope the window could cover the mftr_name part of lookup result. (We vary
	N from 1 to len(mftr_name)). For every step of sliding, We match window with our list of known mftr 
 	until we find a hit (which indicate the current window position cover the mftr_name part of lookup result).
	
*/
const DG_Node* deprecated_findNodeFromName(string& org, const map<string, DG_Node>& name2node){
#ifdef DEBUG
	clog << "#Entered findNodeFromName(), looking for matched DG_Node with org name" << endl;
        clog << "##Input Org Name:" <<  org << endl;
#endif
	static string tmp[] = {	"inc","pte","ltd","co","corp","the","bv","pty","ind", "gmbh","pvt",
        "ing","kg","ohg","sa","srl","bv","llc", "cv", "sl","oy","ab", "ltda", "as","sas","asa","oy","spa"};
	static unordered_set<string> nonMftrWds(tmp, tmp + sizeof(tmp) / sizeof(string)); //known common non-mftr words
	/*
	    legacy non-mtr worrds:
        	"technology","technologies","tech","electronics","electronic";     => Counter-example: Integrated Device Technology, ShengHai Electronics
        	"shenzhen","hangzhou","zhejiang","beijing","shanghai","guangdong"  => Counter-example: Santa Barbara Infrared, ShengHai Electronics
	*/

        /*
                  convert org (the MAC lookup result) to our internal storage form of mftr and mrlv:
                  (1) lower case
                  (2) remove any special characters, leaving only alphabets, numerics and space
                  (3) remove any extra space; i.e. single space only
         */

        str2lower(org); //(1) lower case
        org.erase(remove_if(org.begin(), org.end(), [](char c) { return ! (isalnum(c) || isspace(c)) ; }), org.end()); // (2) leaving only alpha, num, space
        compressSpace(org); //(3) remove any extra space

	/*
		split org string into words and remove non-mftr words
	*/
        istringstream orgWds(org, istringstream::in);  //split org string into words
        string cWd; // tmp string to iterate words in org string
	vector<string> candMftrWds; //candidate mftr wrlds = words in org string - common non-Mftr words;
    	while (orgWds >> cWd) //iterate words from org in cWd
	{	
		if(nonMftrWds.find(cWd)==nonMftrWds.end()){ // if current word not non-mftr, store it
        		candMftrWds.push_back(cWd);
		}
	}

#ifdef DEBUG 
	clog << "##Org Name After Clearning:";
	for (vector<string>::size_type i=0; i!=candMftrWds.size();i++)
	{
		clog << candMftrWds[i] <<" ";
	}	
	clog << endl;
#endif
	//iterate through sub-setences of candMftrWds; e.g. for "a b c" that's "a" "b" "c" "a b" "b c" and "a b c"
	//check if any sub-sentence has a hit in name2node
	for(vector<string>::size_type start=0; start<=candMftrWds.size()-1; start++)
	{
                stringstream sen;
 		for(vector<string>::size_type sen_len=1; start+sen_len-1<=candMftrWds.size()-1; sen_len++)
        	{
			if(!sen.str().empty())
 				sen << " ";
			sen << candMftrWds[start+sen_len-1];
#ifdef DEBUG                    
			clog << "###Testing sub-sen of org name:" << sen.str() << endl;
#endif

			map<string, DG_Node>::const_iterator itr=name2node.find(sen.str());
			if(itr!=name2node.end())
			{
#ifdef DEBUG			
				clog << "####Matched Node found w sub-sen: " << sen.str() << "===> this device is IoT" << endl;
#endif
				return &(itr->second);
			}
		}
	}

	return NULL; //no match is found
}

/*
        output potential mftr related to input node
	by finding all reachable mftr nodes from input node, via BFS
*/
vector<const DG_Node*> findPoMftrsFromNode(const DG_Node& inode){
// Rule:
// if input DG_Node is a mftr; then output list = input node + (if any) reachable mftr nodes from input node
// if input DG_Node is a mrlv; then output list = reachable mftr nodes from input node
// i.e. output list = reachable mftr nodes from input node including itself
#ifdef DEBUG
	clog << "##Searching IoT manufacters reachable from node \"" << inode.getName() << "\"" << endl;
#endif

        vector<const DG_Node*> ret; //ptrs to DG_Node of potential mftrs

	// use BFS to find all reachable
	unordered_set<const DG_Node*> visited; // remember visted nodes to hanlde loop
	queue<const DG_Node*> que;
	que.push(&inode); // insert element at end of queue
	visited.insert(&inode);

	while(!que.empty())
	{
		queue<const DG_Node*>::size_type n=que.size();
		for(queue<const DG_Node*>::size_type ii=0; ii<n; ii++)
		{
			const DG_Node* cnode=que.front();
			que.pop(); //remove front element
		        if(cnode->isMftr())
	                	ret.push_back(cnode);
			const vector<const DG_Node*>& cnode_nexts=cnode->getNexts();
			for(vector<const DG_Node*>::size_type i=0; i!=cnode_nexts.size(); i++)
	       		{
				if(visited.find(cnode_nexts[i])==visited.end()) //if cnode_nexts[i] hasn't been visited before
				{		
					visited.insert(cnode_nexts[i]);
					que.push(cnode_nexts[i]); // insert element at end of queue
				}
       			}

		}	
	}

        return ret;
}


