#include "dg_node.h"
#include "det.h"
#include "iotsteed.h"
#include "dev.h"
#include "sln.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <utility>
#include <algorithm>
#include <arpa/inet.h>
#include <sys/time.h>
#include <pcapplusplus/IPv4Layer.h>
#include <pcapplusplus/IPv6Layer.h>
#include <pcapplusplus/IpAddress.h>
#include <pcapplusplus/EthLayer.h>
#include <pcapplusplus/MacAddress.h>
#include <pcapplusplus/TcpLayer.h>
#include <pcapplusplus/UdpLayer.h>
#include <pcapplusplus/Packet.h>
#include <pcapplusplus/RawPacket.h>
#include "debug.h"

using namespace std;
using namespace pcpp;

/*
	check if a MAC is a device MAC
*/
bool isDevMac(const string& Mac)
{
	bool is_multi_cast=(Mac.rfind("01:00:5e", 0) == 0) || (Mac.rfind("33:33", 0) == 0) || other_multicast_MACs.find(Mac) !=other_multicast_MACs.end(); // if Mac is multicast https://embeddist.wordpress.com/2015/10/07/well-known-ethernet-multicast-address/
	return Mac.compare("ff:ff:ff:ff:ff:ff")!=0 && Mac.compare("00:00:00:00:00:00")!=0 && !is_multi_cast;

}

/*
	count number of occurance of a given character in a string
*/
int count(string s, char c) 
{ 
    int res = 0; 
  
    for (size_t i=0;i<s.length();i++) 
  
        if (s[i] == c) 
            res++; 
    return res; 
} 



/*
	check if an IP is an Unicast IP
*/
bool isUnicast(const string& IP)
{
        int first_oct=stoi(IP.substr(0,IP.find_first_of(".")));
        bool is_multi_cast= (first_oct>=224 && first_oct <=239); // IPv4 muticast range 224.0.0.0 to 239.255.255.255
        return IP.compare("255.255.255.255")!=0 && IP.compare("0.0.0.0")!=0 && !is_multi_cast;
}

/*
	check if an IP is a Public. Private IP includes: 10/8, 172.16.0.0 to 172.31.255.255 , 192.168/16. Also exclude 100.64.0.0/10 reversed by IANA for carrier-grade NAT 
*/
bool isPublic(const string& IP)
{
	int first_oct=stoi(IP.substr(0,IP.find_first_of(".")));
	string last_three_octs=IP.substr(IP.find_first_of(".")+1, string::npos);
	int second_oct=stoi(last_three_octs.substr(0, last_three_octs.find_first_of(".")));
	return !(first_oct==10 || (first_oct==172 && second_oct >=16 && second_oct <=31) || (first_oct==192 && second_oct==168) || (first_oct==100 && second_oct >=64 && second_oct <=127));
}


/*
	check if an IP is public and unicast	

*/
bool isPublicUnicast(const string& IP)
{

	return isUnicast(IP) && isPublic(IP);
}

/*
        is a IP within a /24 prefix
*/
bool withinPrefix(const std::string& IP, const std::string& S24)
{
        return IP.substr(0,IP.find_last_of(".")).compare(S24)==0;
}


/*
	Constructor for IoTSTEED
*/

IoTSTEED::IoTSTEED(const char* mf_fn,const char* key,const char* pref, const char* exc_MACs,  const char* wl_IPs, const char* router_MACs, size_t serverCtThre, const char* bsThre_cstr, const char* outfn_csv):sha_prefix(pref),sha_serverCtThre(serverCtThre),det_apikey(key),def_of(outfn_csv){
        def_of << "pkt_no,timstamp_sec,smac,dmac,sip,dip,sport,dport,proto" << endl;
  	det_name2node=createDGfromCSV(mf_fn);
	string MAC;
	string IP;
	string bsthre;

        stringstream ss(exc_MACs);
	while(getline(ss, MAC, ',')) {
		det_exception_MACs.insert(MAC);
	}

        stringstream ss2(router_MACs);
        while(getline(ss2, MAC, ',')) {
                det_router_MACs.insert(MAC);
        }

        stringstream ss3(bsThre_cstr);
	getline(ss3, bsthre, ',');
	sln_bsThre_abn=(time_t) atoll(bsthre.c_str())*3600;
        getline(ss3, bsthre, ',');
        sln_bsThre_abi=(time_t) atoll(bsthre.c_str())*3600;

        stringstream ss4(wl_IPs);
        while(getline(ss4, IP, ',')) {
		int dot_ct=count(IP,'.');
		if(dot_ct==3)
                {
			sln_whitelisted_IPs.insert(IP);

		} else if (dot_ct==2) {
			sln_whitelisted_s24.insert(IP);
		
		} else if (dot_ct==1) {
			sln_whitelisted_s16.insert(IP);
		} else if (dot_ct==0) {
			sln_whitelisted_s8.insert(IP);
		}
        }

}

/*
        Method that does heavy lifting for IoT detection phase
*/

void IoTSTEED::detect_from_MAC(string devMac, time_t pkt_timestamp){
         string res_mlp=maclookup(devMac, det_apikey);
         const DG_Node* res_nlp=findNodeFromName(res_mlp, det_name2node);

         if(res_nlp==NULL) // no matched node, new device is non-IoT
         {
                 Dev d(devMac,false,sha_serverCtThre,pkt_timestamp);
                 det_mac2dev.insert(pair<string,Dev>(devMac,d));

#if defined(DEBUG) || defined(LOG)
                 clog << "[LOG] "<< pkt_timestamp << ": Detect new non-IoT dev (MAC=" << d.getMAC() <<"; est boot=" << d.getBootTime() << ")"<< endl;;
#endif

         } else { // new device is IoT
                 vector<const DG_Node*> poMftrNds=findPoMftrsFromNode(*res_nlp);
                 Dev d(devMac,true,sha_serverCtThre, poMftrNds,pkt_timestamp);
                 det_mac2dev.insert(pair<string,Dev>(devMac,d));

#if defined(DEBUG) || defined(LOG)
                 clog << "[LOG] "<< pkt_timestamp << ": Detect new IoT dev (MAC="<< d.getMAC() <<"; est boot=" << d.getBootTime() <<"; pot mftrs=";

                 for(vector<const DG_Node*>::const_iterator it=d.getPoMftrNodes()->begin(); it != d.getPoMftrNodes()->end(); it++)
                 {
                         clog << (*it)->getName() << ",";
                 }
                 clog <<")" << endl;
#endif

       }
}


/*
	Meta method that extract packet info and call three phases of IoTSTEED

*/
void IoTSTEED::process_packet(RawPacket& rawPkt, faup_options_t* faup_opts, int pkt_ct){

#ifdef DEBUG
        clog << "\n======= Packet " << pkt_ct << "=====" << endl;
#endif
        Packet parsedPacket(&rawPkt);

        time_t pkt_timestamp=(rawPkt.getPacketTimeStamp()).tv_sec;
        EthLayer* ethLayer = parsedPacket.getLayerOfType<EthLayer>();

        if (ethLayer==NULL) {

#ifdef DEBUG
                clog << "Skipped packet without Ethernet header"<< endl;
#endif
                return;
        }


        MacAddress srcMAC=ethLayer->getSourceMac();
        MacAddress dstMAC=ethLayer->getDestMac();
        string sMac(srcMAC.toString());
        string dMac(dstMAC.toString());
        transform(sMac.begin(), sMac.end(), sMac.begin(), ::tolower);
        transform(dMac.begin(), dMac.end(), dMac.begin(), ::tolower);

#ifdef DEBUG 
        clog << "(Timestamp:" << pkt_timestamp  << ")" << endl;
	clog << "(MAC:" << sMac << "->" << dMac << ")" << endl;
#endif

        if (det_exception_MACs.find(sMac)!=det_exception_MACs.end() || det_exception_MACs.find(dMac)!=det_exception_MACs.end())
        {
#ifdef DEBUG 
                clog << "Ignore Packets from/to Exceptioned Devices" << endl;
#endif
                return;
        }


	/*
		Apply Devce Detection Phase: detect new device from new s/dMac, classify new dev into
					     IoT/non-IoT, if IoT, find potential IoT mftrs
	*/

	apply_detection(pkt_timestamp, sMac, dMac); 


        if (ntohs(ethLayer->getEthHeader()->etherType)!=PCPP_ETHERTYPE_IP) { // if not IPv4 packets

        // IoTSTEED only handle IPv4 traffic for now, The only exception is we check IPv6 DNS traffic since we find some device query A record using IPv6
       // and then talk in IPv4 (Note that for DNS, we track A,AAAA,CNAME type DNS request but only type A and CNAME type reply)

		if (ntohs(ethLayer->getEthHeader()->etherType)==PCPP_ETHERTYPE_IPV6 && parsedPacket.getLayerOfType<IPv6Layer>()!=NULL) // check if IPv6 DNS packets
		{
		        IPv6Layer* ip6Layer = parsedPacket.getLayerOfType<IPv6Layer>();

        		int proto=ip6Layer->getIPv6Header()->nextHeader, sport=-1, dport=-1;

			if (proto==PACKETPP_IPPROTO_UDP && parsedPacket.getLayerOfType<UdpLayer>()!=NULL){ //if IPv6 UDP
            			UdpLayer*  udpLayer = parsedPacket.getLayerOfType<UdpLayer>();
                		sport= ntohs(udpLayer->getUdpHeader()->portSrc);
                		dport= ntohs(udpLayer->getUdpHeader()->portDst);
        		}

      	 		if (proto==PACKETPP_IPPROTO_UDP && (sport ==53 || dport ==53)) //if IPv6 DNS traffic
       			{
#ifdef DEBUG 
				clog << "\nEntered IPv6 DNS Traffic Special Handling" << endl;
#endif
			        bool is_src_IoT=det_mac2dev.find(sMac)!=det_mac2dev.end() && (det_mac2dev.find(sMac)->second).getIsIoT();
			        bool is_dst_IoT=det_mac2dev.find(dMac)!=det_mac2dev.end() && (det_mac2dev.find(dMac)->second).getIsIoT();

               			if(is_src_IoT && dport ==53){ // DNS query from LAN IoT device
#ifdef DEBUG 
		                        clog << "#Parsing IPv6 DNS query from LAN IoT" << endl;
#endif
                        		track_DNS_query(parsedPacket, det_mac2dev.find(sMac)->second, sln_sname2svr, sln_cname2svr, pkt_timestamp, sln_bsThre_abn, sln_bsThre_abi, faup_opts);

               			} else if (is_dst_IoT && sport ==53){ // DNS reply to LAN IoT device

#ifdef DEBUG            
		                        clog << "#Parsing IPv6 DNS reply to LAN IoT" << endl;
#endif
		                        track_DNS_reply(parsedPacket, det_mac2dev.find(dMac)->second, sln_sname2svr, sln_cname2svr, sln_IP2svr,sln_bsThre_abn, sln_bsThre_abi, faup_opts,pkt_timestamp);

               			} else {
#ifdef DEBUG 
		                        clog << "#Ignore IPv6 DNS traffic from/to non-IoT devices" << endl;
#endif
				}

			}  else {

#ifdef DEBUG    
	                        clog << "Skipped non-DNS IPv6 packet" << endl;
#endif
			}
	

		} else {

#ifdef DEBUG    
                	clog << "Skipped non-IPv4-IPv6 packet" << endl;
#endif
		} 

		return;
        }

	// only IPv4 packets goes beyond


        if (parsedPacket.getLayerOfType<IPv4Layer>()==NULL) {

#ifdef DEBUG
                clog << "Skipped IPv4 packet without IPv4 header"<< endl;
#endif
                return;
        }

        if (parsedPacket.getLayerOfType<IPv6Layer>()!=NULL) {

#ifdef DEBUG
                clog << "Skipped IPv4 packet with IPv6 header"<< endl;
#endif
                return;
        }

						   

        IPv4Layer* ip4Layer = parsedPacket.getLayerOfType<IPv4Layer>();
        IPv4Address srcIP = ip4Layer->getSrcIpAddress();
        IPv4Address dstIP = ip4Layer->getDstIpAddress();
        string sIP(srcIP.toString());
        string dIP(dstIP.toString());
        int proto=ip4Layer->getIPv4Header()->protocol, sport=-1, dport=-1;

#ifdef DEBUG 
        clog << "\n(IP: " << sIP << "->" << dIP << "; Proto=" << proto  << ")"<<endl;
#endif

        if(proto==PACKETPP_IPPROTO_TCP)
        {
            TcpLayer*  tcpLayer = parsedPacket.getLayerOfType<TcpLayer>();
            if(tcpLayer==NULL)
            {
#ifdef DEBUG 
                cerr << "[Warning] Malformed TCP packet without TCP header" << endl;
#endif
            } else {
                sport= ntohs(tcpLayer->getTcpHeader()->portSrc);
                dport= ntohs(tcpLayer->getTcpHeader()->portDst);
            }

        } else if (proto==PACKETPP_IPPROTO_UDP){

            UdpLayer*  udpLayer = parsedPacket.getLayerOfType<UdpLayer>();
            if(udpLayer==NULL)
            {
#ifdef DEBUG    
                cerr << "[Warning] Malformed UDP packet without UDP header" << endl;
#endif
            } else {
                sport= ntohs(udpLayer->getUdpHeader()->portSrc);
                dport= ntohs(udpLayer->getUdpHeader()->portDst);
            }
        }

#ifdef DEBUG 
        clog << "(Ports:" << sport <<"->" << dport << ")" << endl;
#endif

        /*
                Apply Server Learning Phase: Track DNS resolution, apply 2-round learning to leanrn legitemate servers IoT devices talk to
        */
	pair<Dev*,string*> p=apply_learning(sport,dport,proto, pkt_timestamp, sMac, dMac, sIP, dIP, rawPkt, parsedPacket, faup_opts);

	/*
		Apply DDoS Defense Phase

	*/
        apply_defense(p.first, p.second, sport, dport, proto, pkt_timestamp, sIP, dIP, pkt_ct, sMac, dMac);

}


/*
	IoT detection phase of IoTSTEED
*/
void IoTSTEED::apply_detection(time_t pkt_timestamp, string& sMac, string& dMac)
{

#ifdef DEBUG 
        clog << "\nEntered IoTSTEED::apply_detection()" << endl;
#endif
	bool unknown_src_dev_mac=det_mac2dev.find(sMac)==det_mac2dev.end() && isDevMac(sMac);
	bool unknown_dst_dev_mac=det_mac2dev.find(dMac)==det_mac2dev.end() && isDevMac(dMac);
	bool is_src_router_mac=det_router_MACs.find(sMac)!=det_router_MACs.end();
	bool is_dst_router_mac=det_router_MACs.find(dMac)!=det_router_MACs.end();

	if(unknown_src_dev_mac && !is_src_router_mac)
	{
#ifdef DEBUG 
		clog << "#Detect device from unknown src MAC" << endl;
#endif
		detect_from_MAC(sMac, pkt_timestamp);
	}

        if(unknown_dst_dev_mac && !is_dst_router_mac)
        {
#ifdef DEBUG 
                clog << "#Detect device to unknown dst MAC" << endl;
#endif
		detect_from_MAC(dMac, pkt_timestamp);		
        }

#ifdef DEBUG
	if(!(unknown_src_dev_mac && !is_src_router_mac) && !(unknown_dst_dev_mac && !is_dst_router_mac)){
               clog <<"#Ignore packets with Macs that are known or not from devices or from routers" << endl;
       }
#endif


}

/*
	Server learning phase of IoTSTEED
*/
pair<Dev*,string*> IoTSTEED::apply_learning(int sport, int dport, int proto, time_t pkt_timestamp, string& sMac, string& dMac, string& sIP, string& dIP, RawPacket& rawPkt, Packet& parsedPacket, faup_options_t *faup_opts)
{
#ifdef DEBUG 
	clog << "\nEntered IoTSTEED::apply_learning()" << endl;
#endif

        bool is_src_LAN=withinPrefix(sIP, sha_prefix);
        bool is_dst_LAN=withinPrefix(dIP, sha_prefix);
        bool is_src_public_unicast=isPublicUnicast(sIP);
        bool is_dst_public_unicast=isPublicUnicast(dIP);
        bool is_src_IoT=det_mac2dev.find(sMac)!=det_mac2dev.end() && (det_mac2dev.find(sMac)->second).getIsIoT();
        bool is_dst_IoT=det_mac2dev.find(dMac)!=det_mac2dev.end() && (det_mac2dev.find(dMac)->second).getIsIoT();

        /*
		Two type of servers : server accssed by names (actual server accessing preceded by DNS query fr server name), short as abn-servers
				      server accssed by IPs (server directly acccessed by IP without preceding DNS queries), short as abi-servers
	*/

	/*
		Apply Server Learning to LAN IoT's traffic

                For DNS Traffic:
                	If DNS Query from LAN IoT:          (1) Create a Svr obj for every abn-servers (by identifying every queried, distinct sname)
						            (2) Record a list of abn-servers Svr objs that each IoT query/will talk to
			 	                            (3) 2-round learning of abn-servers for each LAN IoT querying them

			If DNS Reply to LAN IoT:            (1) Extract IPs and CNames for abn-servers Svr objs created above

                For all LAN-IoT-WAN-Server traffic:         (1) Crate a Svr obj for every abi-servers
		(including trafic to WAN DNS server IP)	    (2) Record a list of abi-servers Svr objts that each IoT talk to
 						            (3) 2-round learning of abi-servers that LAN IoT talk to
        */


       if (proto==PACKETPP_IPPROTO_UDP && (sport ==53 || dport ==53)) //DNS traffic
       {
               if(is_src_LAN && is_src_IoT && dport ==53){ // DNS query from LAN IoT device
#ifdef DEBUG 
               		clog << "#Parsing DNS query from LAN IoT" << endl;
#endif
               		track_DNS_query(parsedPacket, det_mac2dev.find(sMac)->second, sln_sname2svr, sln_cname2svr, pkt_timestamp, sln_bsThre_abn, sln_bsThre_abi, faup_opts);

               } else if (is_dst_LAN && is_dst_IoT && sport ==53){ // DNS reply to LAN IoT device
#ifdef DEBUG            
                        clog << "#Parsing DNS reply to LAN IoT" << endl;
#endif
               		track_DNS_reply(parsedPacket, det_mac2dev.find(dMac)->second, sln_sname2svr, sln_cname2svr, sln_IP2svr, sln_bsThre_abn, sln_bsThre_abi, faup_opts, pkt_timestamp);

               } else {
#ifdef DEBUG 
                        clog << "#Ignore DNS traffic from/to non-IoT devices" << endl;
#endif
		}	

        }

        Dev* lan_iot=NULL;
        string* wan_svr_IP=NULL;


	if (is_src_LAN && is_src_IoT && !is_dst_LAN && is_dst_public_unicast){ // traffic from LAN IoT to WAN server (a unicast WAN IP)
#ifdef DEBUG 
                clog << "#Handling Traffic from LAN IoT to WAN servers" << endl;
#endif
		lan_iot=&(det_mac2dev.find(sMac)->second);
		wan_svr_IP=&dIP;

	} else if (is_dst_LAN && is_dst_IoT && !is_src_LAN && is_src_public_unicast) { // traffic from WAN server (a unicast WAN IP) to LAN IoT 
#ifdef DEBUG 
                clog << "#Handling Traffic from WAN servers to LAN IoT" << endl;
#endif
		lan_iot=&(det_mac2dev.find(dMac)->second);
		wan_svr_IP=&sIP;
	} else { // skip other traffic
#ifdef DEBUG 
                clog << "#Traffic not between LAN IoT and WAN servers: do nothing" << endl;
#endif
		return make_pair(lan_iot, wan_svr_IP);
	}

	handle_lan_iot_wan_server_traffic(lan_iot, wan_svr_IP, sln_whitelisted_IPs, sln_whitelisted_s24,sln_whitelisted_s16,sln_whitelisted_s8, sln_sname2svr, sln_IP2svr, pkt_timestamp, sln_bsThre_abn, sln_bsThre_abi,faup_opts);

	return make_pair(lan_iot, wan_svr_IP);
}


/*
	DDoS defense phase of IoTSTEED
*/
void IoTSTEED::apply_defense(Dev* lan_iot, string* wan_svr_IP, int sport, int dport, int proto, time_t pkt_timestamp, string& sIP, string& dIP, int pkt_ct, string& sMac, string& dMac)
{

#ifdef DEBUG 
	cout << "\nEntered IoTSTEED::apply_defense()" << endl;
#endif
	if(lan_iot!=NULL && wan_svr_IP!=NULL) //traffic between LAN IoT and WAN Server (otherwise both will be NULL)
	{
		Svr* wan_svr;
		bool pass;

		if(sln_IP2svr.find(*wan_svr_IP)!=sln_IP2svr.end()){ // abn-server

			wan_svr=(sln_IP2svr.find(*wan_svr_IP))->second;

		} else if(sln_sname2svr.find(*wan_svr_IP)!=sln_sname2svr.end()){ //abi-server

			wan_svr=&((sln_sname2svr.find(*wan_svr_IP))->second);
			if(!wan_svr->abn && !lan_iot->abi_def) //pass all abi servers for devices that we turn off abi defense
			{
#ifdef DEBUG 
				cout << "#DECESION: PASS traffic to/from ABI server for devices that we turn off ABI defense" << endl;
#endif
			        return;
			}
			
		} else {

			cerr << "[Error] server IP " << *wan_svr_IP << "not mathing known abn- and abi-servers" << endl;
			return;
		}

		if((wan_svr->dev2svr_cls).find(lan_iot)!=(wan_svr->dev2svr_cls).end()){



			pass=((wan_svr->dev2svr_cls).find(lan_iot))->second;
			if(pass) {
#ifdef DEBUG 
		        	cout << "#DECESION: PASS traffic to/from IoT server" << endl;
#endif
			} else {
#ifdef DEBUG 
                                cout << "#DECESION: DROP traffic to/from IoT server" << endl;
#endif
				def_of << pkt_ct << "," << pkt_timestamp << "," << sMac << "," << dMac << "," << sIP << "," << dIP << "," << sport << "," << dport << "," << proto << endl;

			}
#ifdef DEBUG 
                        cout << "##LAN IoT Dev=" << lan_iot->MAC << "; WAN Svr=" << wan_svr->sname << endl;
#endif

		} else {

			cerr << "[Error] No classification for device " << lan_iot->MAC << " at server " << wan_svr->sname << endl;
			return;
		}
		
	} else {
#ifdef DEBUG 
        	cout << "#Traffic not between LAN IoT and WAN servers: do nothing" << endl;
#endif
	}
}


