#include "dev.h"
#include "dg_node.h"
#include "det.h"
#include "iotsteed.h"
#include "version.h"
#include <iostream>
#include <string>
#include <sstream>
#include <getopt.h>
#include <arpa/inet.h>
#include <pcapplusplus/EthLayer.h>
#include <pcapplusplus/IPv4Layer.h>
#include <pcapplusplus/IPv6Layer.h>
#include <pcapplusplus/Packet.h>
#include <pcapplusplus/PcapFileDevice.h>
#include <pcapplusplus/PcapLiveDeviceList.h>
#include <faup/faup.h>
#include <faup/options.h>
#include <faup/output.h>
#include "debug.h"
#include <signal.h>


using namespace std;
using namespace pcpp;

void usage(){
        cerr<<"Usage:\n"
	"--------\n"
	" iotsteed -i/I input_pcap/interface_ip -m mftr_mrlv_list -k oui_lookup_key -o out_csv -r router_MACs [-e MAC_exceptions] [-w whitelisted_IPs] [-p network_prefix ] [-s svr_count_thre] [-b bootstrapping_thre]\n"
	"\n"
        " -i/--input-pcap              	: Input pcap file that iotsteed will read packet from and operate on (offline mode)\n"
        " -I/--Input-interface           : Interface IP that iotsteed will read packet from and operate on (online mode)\n"
        " -m/--mapping	        	: The csv file storing mapping between known IoT manufacuter (mftr) and IoT manufacuter relative (mrlv).\n"
	"			          (special mrlv: - means a mftr has no mrlv; null means a mftr has special non-existent mrlv) \n"
        " -k/--key                       : API key for OUI lookup at macaddress.io\n"
	" -o/--output-prefix		: prefix for output files for (1) device detected (2) server learnt (3) ddos packets dropped\n"
        " -r/--router                    : A list of comma-seperated MACs for LAN routers. We do not apply detection and learning to LAN routers.\n"
        " -e/--exception          	: A list of comma-seperated MACs for devices whose packets we ignore.\n"
        " -w/--whitelisted_IPs         	: A list of comma-seperated whitelisted server IP and prefixes: IP like 1.2.3.4, /24 prefix like 1.2.3, /16 prefix 1.2 and /8 prefix like 1.\n"
	"				  Default is '8.8.8.8,8.8.4.4'\n"
        " -p/--prefix                  	: /24 prefix for the deployed access network (in format 1.2.3). Default is 192.168.1. \n"
	" -s/--server-count-threshold  	: Max number of distinct servers an IoT device should talk to. Default is 70\n"
        " -b/--bootstrapping-threshold   : Duration of server bootstrapping (unit: hours). Default is 2,120 (2 hours for ABN and 120 hours for ABI servers).\n"
        "--------\n"
        "Version: " << VERSION << "\n"
	"Debugging " 
	#ifdef DEBUG 
	"On" 
	#else 
	"Off" 
	#endif
	"\n"
        "Logging   "
        #ifdef LOG
        "On"
        #else
        "Off"
        #endif
        "\n";
        exit(1);
}

struct cookie_struct{
        IoTSTEED* isd;
        faup_options_t* faup_opts;
        int pkt_ct;
};

volatile sig_atomic_t flag = 0;

void sig_handler(int sig){
	flag=1;
}
/**
* a callback function for the blocking mode capture which is called each time a packet is captured
*/
static bool onPacketArrivesBlockingMode(RawPacket* rawPkt, PcapLiveDevice* dev, void* cookie)
{

    signal(SIGINT, sig_handler);
    cookie_struct* cks = (cookie_struct*) cookie;
    cks->pkt_ct=cks->pkt_ct+1;    
    cks->isd->process_packet(*rawPkt,cks->faup_opts,cks->pkt_ct);

    // return false means we don't want to stop capturing after this callback
    if(flag){
	clog << endl << "Ctrl+C Detected; Outputing Results" << endl;
    }
    return flag;
}


int main(int argc, char* argv[]){
	const char * infn_pcap="";        //one of infn_pcap and interface_IP is required
        const char * interface_IP="";      //one of infn_pcap and interface_IP is required
	const char * outprefix_cstr="";        //mandatory
	const char * fn_mftrmrlv="";    //mandatory
	const char * mac_apikey="";     //mandatory 
        const char * router_MACs="";  //mandatory
        const char * whitelisted_IPs="8.8.8.8,8.8.4.4";  //optional with default
        const char * exception_MACs=""; //optional without default 
	const char * prefix="192.168.1";  //optinal with default
	const char * svrCtThre_cstr="70";   //optinal with default 
        const char * bsThre_cstr="2,120";   //optinal with default 
	int opt;

	static struct option long_options[] =
	{
	  {"help",           no_argument,       0, 'h'},
	  {"input",     required_argument, 0,  'i'},
	  {"interface",     required_argument, 0,  'I'},
          {"output-prefix",     required_argument, 0,  'o'},
          {"mapping", required_argument, 0,  'm'},
          {"key",     required_argument, 0,  'k'},
	  {"prefix",         required_argument, 0,  'p'},
	  {"router",  required_argument, 0,  'r'},
	  {"exception",  required_argument, 0,  'e'},
	  {"whitelisted_IPs",  required_argument, 0,  'w'},
	  {"server-count-threshold", required_argument, 0,  's'}, 
	  {"bootstrapping-threshold", required_argument, 0,  'b'}, 
	  {0 ,0, 0, 0}
	};

	while((opt = getopt_long( argc, argv, "hi:I:o:m:k:p:r:e:w:s:b:", long_options,NULL))!=-1 ){
		switch (opt)
        	{
        		case 0: break;
  
 			case 'i':
				infn_pcap=optarg;
				break;
                        case 'I':
                                interface_IP=optarg;
                                break;
                        case 'o':
                                outprefix_cstr=optarg;
                                break;
			case 'm':
				fn_mftrmrlv=optarg;
				break;
			case 'k':
				mac_apikey=optarg;
				break;
			case 'p':
				prefix=optarg;
				break;
                        case 'r':
                                router_MACs=optarg;
                                break;
			case 'e':
				exception_MACs=optarg;
				break;
                        case 'w':
                                whitelisted_IPs=optarg;
                                break;
			case 's':
				svrCtThre_cstr=optarg;
				break;
                        case 'b':
                                bsThre_cstr=optarg;
                                break;
			case '?':
				break;
	                default:
        	                usage();
		}
	}

	if ((infn_pcap[0]=='\0' && interface_IP[0]=='\0') || (infn_pcap[0]!='\0' && interface_IP[0]!='\0') || outprefix_cstr[0]=='\0' ||  fn_mftrmrlv[0]=='\0' || mac_apikey[0]=='\0' || router_MACs[0]=='\0') //ensure mandatory parameters are supplied
	{
		usage();	
		return 1;
	}

    size_t svrCtThre=0;
    sscanf(svrCtThre_cstr, "%zu", &svrCtThre);
    string outprefix(outprefix_cstr);
    string out_ddt, out_sln, out_def;
    out_ddt=outprefix+"_ddt.csv";
    out_sln=outprefix+"_sln.csv";
    out_def=outprefix+"_def.csv";

    ofstream ddt_of(out_ddt);
    ofstream sln_of(out_sln);

    // initilize an IoTSTEED object which keeps track of states of defense 
    // and applies our three-phase processing to each packet
    IoTSTEED* isd= new IoTSTEED(fn_mftrmrlv,mac_apikey,prefix,exception_MACs,whitelisted_IPs,router_MACs,svrCtThre,bsThre_cstr,out_def.c_str());

    // Prepare server domain extracter faup
    faup_options_t *faup_opts;
    faup_opts = faup_options_new();
    faup_opts->output = FAUP_OUTPUT_NONE;
    faup_opts->exec_modules = FAUP_MODULES_NOEXEC;


    clog << "### IoTSTEED Version " << VERSION <<  " ###" << endl;
    if(infn_pcap[0]!='\0'){
    	clog << "Offline Mode: reading packets from " << infn_pcap << endl;
    	PcapFileReaderDevice pcap_reader(infn_pcap);
    	if (!pcap_reader.open())
    	{
    		cerr << "[Error] Can't open input pcap file\n" << endl;
    		return 1;
    	}

    	RawPacket rawPkt; 
	int pkt_ct=0;

    	while (pcap_reader.getNextPacket(rawPkt))
    	{
    	    pkt_ct=pkt_ct+1;
    	    isd->process_packet(rawPkt,faup_opts,pkt_ct);
    	}

    } else if (interface_IP[0]!='\0'){
	clog << "Online Mode: reading packets from " << interface_IP << "; End packet reading and output result by ctrl+c;" << endl;

	PcapLiveDevice* dev = PcapLiveDeviceList::getInstance().getPcapLiveDeviceByIp(interface_IP);
	if (dev == NULL)
	{
	    cerr << "Cannot find interface with IPv4 address of " << interface_IP << endl;
	    exit(1);
	}
	if (!dev->open())
	{
	    cerr << "Cannot open device\n" << endl;
	    exit(1);
	}

    	struct cookie_struct cks;
    	cks.isd=isd;
    	cks.faup_opts=faup_opts;
	cks.pkt_ct=0;

	dev->startCaptureBlockingMode(onPacketArrivesBlockingMode, &cks, 0);
	// thread is blocked until capture is finished
	// After capture is finished, output results
      
   }
    //output devices learnt and servers they talk to
    map<string, Dev>& mac2dev=isd->det_mac2dev;
    ddt_of << "mac,dev_cls,boot_time,po_mftrs,total_svr_ct,abn_svr_ct,abi_svr_ct,nonwl_abi_svr_ct,bs_abi_rate,max_post_bs_abi_rate,abi_turnoff_trial,abi_def" << endl;
    sln_of << "mac,svr_sname,svr_cls,frist_access_time,learning_round,svr_access_type" << endl;
    
    for(map<string, Dev>::iterator it=mac2dev.begin(); it!=mac2dev.end(); it++){
	Dev& cdev=it->second;
	// output info for this learnt device
	ddt_of << cdev.getMAC() << ",";

	if(cdev.getIsIoT()) {
		ddt_of << "IoT,";	
	} else {
		ddt_of << "nonIoT,";
	}

       	ddt_of << cdev.getBootTime() << ",";

        const vector<const DG_Node*>& poMftrs=cdev.poMftrNodes;

        for( vector<const DG_Node*>::size_type i=0; i<poMftrs.size(); i++)
        {
		ddt_of << poMftrs[i]->getName();
		if(i!=poMftrs.size()-1) 
			ddt_of <<"|";
	}

	ddt_of << "," << cdev.getServerCt() << "," << cdev.getAbnServerCt() << ",";
	ddt_of <<  cdev.abi_servers.size() << "," << cdev.abi_servers.size()-cdev.wl_abi_servers.size() << "," << cdev.bs_abi_rate << "," << cdev.max_post_bs_abi_rate << "," << cdev.abi_turnoff_trial << ",";
	if(cdev.abi_def)
	{
		ddt_of << "on" << endl;
	} else {

		ddt_of << "off" << endl;
	}

        //output server leanrt for each device by iterating 
	const unordered_set<Svr*>& svrs=cdev.servers;
	for(unordered_set<Svr*>::const_iterator is=svrs.begin(); is!=svrs.end(); is++)
	{

		Svr* csvr=*is;
		sln_of <<  cdev.getMAC() << "," << csvr->sname << ",";
		if( (csvr->dev2svr_cls).find(&cdev) != (csvr->dev2svr_cls).end()){
			if(((csvr->dev2svr_cls).find(&cdev))->second)
				sln_of << "benign";
			else
				sln_of << "malicious";
		} else {
				sln_of << "N/A";
		}

		sln_of <<",";

                if( (csvr->dev2first_access_time).find(&cdev) != (csvr->dev2first_access_time).end()){
                                sln_of << ((csvr->dev2first_access_time).find(&cdev))->second;
                } else {
                                sln_of << "N/A,";
                }

                sln_of <<",";

                if( (csvr->dev2rounds).find(&cdev) != (csvr->dev2rounds).end()){
                                sln_of << ((csvr->dev2rounds).find(&cdev))->second;
                } else {
                                sln_of << "N/A,";
                }

                sln_of <<",";

                if(csvr->abn) 
			sln_of << "abn";
		else
			sln_of << "abi";

		sln_of << endl;
	}

   }
    
    delete isd;

}

